#!/usr/bin/env Rscript --vanilla

# Plot coverage matrix generated by e.g. 'mccortex31 links --covg-hist out.csv ...'
#
args <- commandArgs(trailingOnly=TRUE)
if((length(args) < 2 || length(args) > 3) && (length(args) < 6 || length(args) > 8)) {
  stop("Usage: ./link-cov-heatmap.R <covg.csv> <out.pdf> [<cutoff> [<k> <kcov> <readlen> [maxcov [maxdist]]]]\n")
}

cutoff=0
kmer=0
kcov=0
readlen=0
maxcov=0
maxdist=0

input_csv <- args[1]
output_pdf <- args[2]
if(length(args) >= 3) { cutoff <- as.numeric(args[3]) }
if(length(args) >= 4) { kmer <- as.numeric(args[4]) }
if(length(args) >= 5) { kcov <- as.numeric(args[5]) }
if(length(args) >= 6) { readlen <- as.numeric(args[6]) }
if(length(args) >= 7) { maxcov <- as.numeric(args[7]) }
if(length(args) >= 8) { maxdist <- as.numeric(args[8]) }

library('ggplot2')
library('reshape')
library('scales')
library('plyr')

cat("input_csv='",input_csv,"'\n",sep='')
cat("output_pdf='",output_pdf,"'\n",sep='')
cat('cutoff=',cutoff,'\n',sep='')
cat('kmer=',kmer,'\n',sep='')
cat('kcov=',kcov,'\n',sep='')
cat('readlen=',readlen,'\n',sep='')
cat('maxcov=',maxcov,'\n',sep='')
cat('maxdist=',maxdist,'\n',sep='')

r <- read.table(input_csv,sep=',',head=T,row.names=1,comment.char='#')

if(maxcov == 0) { maxcov=ncol(r) }
if(maxdist == 0) { maxdist=nrow(r) }
maxcov<-min(ncol(r), maxcov)
maxdist<-min(nrow(r), maxdist)

# Take subset
if(ncol(r) > maxcov) {
  for(i in (maxcov+1):ncol(r)) { r[,maxcov] <- r[,maxcov]+r[,i] }
  r <- r[,1:maxcov]
}

if(nrow(r) > maxdist) {
  r<-r[1:maxdist,]
}

# Convert row/column names to numbers
colnames(r) <- 1:ncol(r)
rownames(r) <- 1:nrow(r)

# Normalise columns
norm <- function(x) {
  max = max(x); min = min(x); diff = max(x)-min(x)
  if(diff == 0) { return (x); }
  else { return((x-min) / diff); }
}

r.m <- t(apply(as.matrix(r), 1, norm))
d.m <- melt(r.m, varnames=c("dist","cov"))
d.m <- ddply(d.m, .(dist), transform)

get_step <- function(max) {
  s <- floor((max+99)/100)*10
  if(s == 0) { return(1); }
  else { return(s); }
}

xstep<-get_step(maxdist)
xticks=seq(xstep,maxdist,xstep)
ystep<-get_step(maxcov)
yticks=seq(ystep,maxcov,ystep)

hist.median <- function(x) {
  s<-sum(x)/2
  for(i in 1:length(x)) {
    s<-s-x[i]
    if(s <= 0) { return(i) }
  }
  return(length(x));
}

medians <- apply(r, 1, hist.median)
lines<-data.frame(x=1:maxdist, y=medians, lines=rep("avgmed",maxdist))

nlines <- 1
lines.leg.breaks <- c("avgmed")
lines.leg.labels <- c("avgmed"="Median")
lines.leg.ltypes <- c("avgmed"="solid")
lines.leg.colors <- c("avgmed"="firebrick")

if(cutoff > 0) {
  nlines <- nlines+1
  lines.leg.breaks <- c(lines.leg.breaks, "cutoff")
  lines.leg.labels <- c(lines.leg.labels, "cutoff"="Cleaning cut-off")
  lines.leg.ltypes <- c(lines.leg.ltypes, "cutoff"="solid")
  lines.leg.colors <- c(lines.leg.colors, "cutoff"="black")

  lines<-rbind(lines, data.frame(x=1:maxdist, y=rep(cutoff, maxdist),
                                 lines=rep("cutoff",maxdist)))
}

if(kmer > 0 && kcov > 0 && readlen > 0) {
  # Plot expected coverage
  # kcov = cov * (readlen-k+1) / readlen
  cov = kcov * readlen / (readlen-kmer+1)
  x <- 1:maxdist
  expcov <- cov * (readlen - (kmer+1:maxdist) + 1) / readlen;
  expcov[expcov<0] <- 0 # set negative values to zero

  nlines <- nlines+1
  lines.leg.breaks <- c(lines.leg.breaks, "expcov")
  lines.leg.labels <- c(lines.leg.labels, "expcov"="Expected")
  lines.leg.ltypes <- c(lines.leg.ltypes, "expcov"="dashed")
  lines.leg.colors <- c(lines.leg.colors, "expcov"="black")

  lines<-rbind(lines, data.frame(x=1:maxdist, y=expcov, lines=rep("expcov",maxdist)))
}

p <- ggplot(d.m, aes(dist, cov)) + theme(panel.background = element_rect(fill = 'white', colour = 'white')) +
     geom_tile(aes(fill = value), color = "white") +
     scale_fill_gradient(low = "white", high = "steelblue") +
     ggtitle("Link coverage") +
     scale_x_discrete(name="Distance (kmers)", breaks=rownames(r)[xticks], labels=xticks) +
     scale_y_discrete(name="Coverage", breaks=colnames(r)[yticks], labels=yticks) +
     geom_line(data=lines, aes(x=x, y=y, col=lines, linetype=lines)) +
     scale_linetype_manual(name="Coverage",
                           breaks=lines.leg.breaks,
                           labels=lines.leg.labels,
                           values=lines.leg.ltypes) +
     scale_color_manual(name="Coverage",
                        breaks=lines.leg.breaks,
                        labels=lines.leg.labels,
                        values=lines.leg.colors)

ggsave(p, file=output_pdf, width=6, height=6)
